/*
 * 2D Game Engine 
 * ECS.h:
 * Based on pikuma.com 2D game engine in C++ and Lua course
 * Copyright (c) 2021 986-Studio. All rights reserved.
 *
 * Created by Manoël Trapier on 11/02/2021.
 */

#ifndef GAMEENGINE_ECS_H
#define GAMEENGINE_ECS_H

#include <stdint.h>

#include <bitset>
#include <vector>
#include <unordered_map>
#include <typeindex>
#include <set>

#include <Pool.h>

const uint32_t MAX_COMPONENTS = 32;

typedef std::bitset<MAX_COMPONENTS> Signature;

struct IComponent
{
protected:
    static uint32_t nextId;
};

template <typename T> class Component: public IComponent
{
public:
    static uint32_t getId()
    {
        static auto id = IComponent::nextId++;
        return id;
    }
};

class Entity
{
private:
    uint32_t id;

public:
    explicit Entity(uint32_t id): id(id) {};
    Entity(const Entity & entity) = default;

    uint32_t getId() const;

    Entity& operator=(const Entity & other) = default;
    bool operator==(const Entity &other) const { return this->id == other.id; };
    bool operator!=(const Entity &other) const { return this->id != other.id; };
    bool operator>(const Entity &other) const  { return this->id > other.id; };
    bool operator<(const Entity &other) const  { return this->id < other.id; };
};

class System
{
private:
    Signature componentSignature;
    std::vector<Entity> entities;

public:
    System() = default;
    ~System() = default;

    void addEntity(Entity entity);
    void removeEntity(Entity entity);
    std::vector<Entity> getEntities() const;
    const Signature& getComponentSignature() const;

    template<typename T> void requireComponent();
};

template<typename T> void System::requireComponent()
{
    const auto componentId = Component<T>::getId();
    this->componentSignature.set(componentId);
}

class Registry
{
private:
    uint32_t numEntities = 0;

    std::vector<IPool*> componentPools;
    std::vector<Signature> entityComponentSignatures;

    std::unordered_map<std::type_index, System*> systems;

    std::set<Entity> entitiesToBeAdded;
    std::set<Entity> entitiesToBeKilled;

public:
    Registry() = default;
    ~Registry() = default;

    /* Entities */
    Entity CreateEntity();
    void addEntityToSystem(Entity entity);

    /* Components */
    template <typename T, typename ...TArgs> void addComponent(Entity entity, TArgs&& ...args);
    template <typename T> void removeComponent(Entity entity);
    template <typename T> bool hasComponent(Entity entity);

    /* Systems */
    template<typename T, typename ...TArgs> void addSystem(TArgs&& ...args);
    template<typename T> void removeSystem();
    template<typename T> bool hasSystem() const;
    template<typename T> T& getSystem() const;

    /* Others */
    void Update();
};

template<typename T, typename ...TArgs> void Registry::addComponent(Entity entity, TArgs&& ...args)
{
    const auto componentId = Component<T>::getId();
    const auto entityId = entity.getId();
    Pool<T> *componentPool;
    /* Check the pool is big enough */
    if (componentId >= this->componentPools.size())
    {
        this->componentPools.reserve(componentId + 1, nullptr);
    }

    /* Is this component exist in the pool? */
    if (!this->componentPools[componentId])
    {
        this->componentPools[componentId] = new Pool<T>();
    }

    componentPool = this->componentPools[componentId];

    if (entityId >= componentPool->getSize())
    {
        componentPool->resize(this->numEntities);
    }

    T newComponent(std::forward<TArgs>(args)...);

    componentPool->set(entityId, newComponent);
    entityComponentSignatures[entityId].set(componentId);
}

template <typename T> void Registry::removeComponent(Entity entity)
{
    const auto componentId = Component<T>::getId();
    const auto entityId = entity.getId();

    entityComponentSignatures[entityId].set(componentId, false);
}

template <typename T> bool Registry::hasComponent(Entity entity)
{
    const auto componentId = Component<T>::getId();
    const auto entityId = entity.getId();

    return entityComponentSignatures[entityId].test(componentId);
}

template<typename T, typename ...TArgs> void Registry::addSystem(TArgs&& ...args)
{
    T *newSystem(new T(std::forward<TArgs>(args)...));

    this->systems.insert(std::make_pair(std::type_index(typeid(T)), newSystem));
}

template<typename T> void Registry::removeSystem()
{
    auto system = this->systems.find(std::type_index(typeid(T)));
    this->systems.erase(system);
}

template<typename T> bool Registry::hasSystem() const
{
    return this->systems.find(std::type_index(typeid(T))) != this->systems.end();
}

template<typename T> T& Registry::getSystem() const
{
    auto system = this->systems.find(std::type_index(typeid(T)));
    return *(std::static_pointer_cast<T>(system->second));
}

#endif /* GAMEENGINE_ECS_H */
